/*
 * top - a top users display for Unix
 *
 * SYNOPSIS:  Any Sun running SunOS 5.7+ (Solaris 2.7+)
 *
 * DESCRIPTION:
 * This is the machine-dependent module for SunOS 5.7+ (Solaris 2.7+).
 *
 * This makes top work on the following systems:
 *         SunOS 5.6 (maybe - not tested)
 *         SunOS 5.7
 *         SunOS 5.8
 *         SunOS 5.9
 *
 * LIBS: -lelf -lkstat
 *
 * CFLAGS: -DHAVE_GETOPT -DORDER -DHAVE_STRERROR
 *
 * AUTHOR:       Chris Hughes           <chris@spang.uk.eu.org>
 * CONTRIBUTORS: Erik OShaughnessy      <eriko@schwa.Central.Sun.COM>
 *
 * With code and concepts from original sunos5 port by:
 *
 * AUTHORS:      Torsten Kasch          <torsten@techfak.uni-bielefeld.de>
 *               Robert Boucher         <boucher@sofkin.ca>
 * CONTRIBUTORS: Marc Cohen             <marc@aai.com>
 *               Charles Hedrick        <hedrick@geneva.rutgers.edu>
 *               William L. Jones       <jones@chpc>
 *               Petri Kutvonen         <kutvonen@cs.helsinki.fi>
 *               Casper Dik             <casper.dik@sun.com>
 *               Tim Pugh               <tpugh@oce.orst.edu>
 */


#include "top.h"
#include "machine.h"
#include "utils.h"

#include "colour.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <unistd.h>
#include <procfs.h>
#include <sys/proc.h>
#include <sys/param.h>
#include <kstat.h>
#include <sys/sysinfo.h>
#include <sys/swap.h>
#include <vm/anon.h>

#define PROCFS          "/proc"

/* Useful externs */
extern char *myname;
extern void quit(int);

/*
 * Process information built up on each run
 */
typedef struct proc_info {
  /* From psinfo */
  int            pr_nlwp;          /* number of lwps in process */
  pid_t          pr_pid;           /* Process id */
  uid_t          pr_uid;           /* Real user id */
  size_t         pr_size;          /* Size of process image in Kbytes */
  size_t         pr_rssize;        /* resident set size in Kbytes */
  ushort_t       pr_pctcpu;        /* % of recent cpu time used by all lwps */
  /*  ushort_t       pr_pctmem;        * % of system memory used by process */
  timestruc_t    pr_time;          /* usr+sys cpu time for this process */
  char           pr_fname[PRFNSZ]; /* name of execed file */
  /* From psinfo.pr_lwp (representative lwp) */
  processorid_t  pr_onpro;         /* processor which last ran the lwp */
  int            pr_pri;           /* priority, high value is high priority */
  char           pr_nice;          /* nice for cpu usage */
  char           pr_state;         /* numeric lwp state */
  /* From prusage */
  ulong_t        pr_majf;          /* major page faults */
  ulong_t        pr_majf_abs;      /* Absolute majf for process */
} proc_info_t;

typedef struct old_proc_info {
  uchar_t        pr_valid;         /* Is this entry in use (1/0) */
  pid_t          pr_pid;
  ulong_t        pr_majf_abs;
} old_proc_info_t;

typedef struct handle {
  int current_index;
} handle_t;

/*
 *  These definitions control the format of the per-process area
 */
static char header[] =
"   PID X        THR PR NCE  SIZE   RES STATE   TIME FLTS    CPU COMMAND";
/* 0123456   -- field to fill in starts at header+6 */
#define UNAME_START 7
#ifdef COLOUR
#define Proc_format "%6d %-8.8s %3d %2d %3d %s%5s%s %s%5s%s %-5s %s%6s%s %s%4lu%s %s%5.2f%%%s %s"
#else
#define Proc_format "%6d %-8.8s %3d %2d %3d %5s %5s %-5s %6s %4lu %5.2f%% %s"
#endif

/* process state names for the "STATE" column of the display */
char *state_abbrev[] =
  { "", "sleep", "run", "zombi", "stop", "start", "cpu", "swap" };

int process_states[8];
char *procstatenames[] =
  { "", " sleeping, ", " running, ", " zombie, ", " stopped, ",
    " starting, ", " on cpu, ", " swapped, ",  NULL };

#define CPUSTATES       5
int cpu_states[CPUSTATES];
char *cpustatenames[] =
  { "idle", "user", "kernel", "iowait", "swap", NULL };

/* these are for detailing the memory statistics */
int memory_stats[5];
char *memorynames[] =
  { "K real, ", "K free, ", "K swap in use, ", "K swap free", NULL };

/* these are names given to allowed sorting orders -- first is default */
char *ordernames[] = 
  { "cpu", "size", "res", "time", "flts", NULL };

#ifdef COLOUR
char *colournames[] =
  { "host", "1min", "5min", "15min", "cpu", "size", "res", "time", "flts", NULL };
#endif

/* forward definitions for comparison functions */
int compare_cpu();
int compare_size();
int compare_res();
int compare_time();
int compare_flts();

int (*proc_compares[])() = {
    compare_cpu,
    compare_size,
    compare_res,
    compare_time,
    compare_flts,
    NULL };

/* We hold /proc open all the time */
static DIR *procdir;

/* Size of the process arrays */
static unsigned int process_array_size = 0;
/* Storage for info on each process */
static proc_info_t *process_array = NULL;
/* Storage for info that should be saved between runs */
static old_proc_info_t *old_process_array = NULL;

/* Keep a link to the number of CPUs this machine has. */
static int ncpus = 0;
   
/*
 * The arrays for holding process information are defined on
 * the first call to machine_init (which gets the no of procs
 * from kstat and inits to that), and then from get_process_info,
 * where it will increment slighly each time
 */
#define REALLOC_INCREMENT  10
static void 
realloc_arrays (int nsize)
{
  nsize += REALLOC_INCREMENT;
  if (nsize <= process_array_size)
    return;

  process_array_size = nsize;
  process_array = (proc_info_t *)realloc(process_array,
					   (nsize*sizeof(proc_info_t)));
  old_process_array = (old_proc_info_t *)realloc(old_process_array,
						 (nsize*sizeof(old_proc_info_t)));

  if (!process_array || !old_process_array) {
    fprintf (stderr,"%s: can't allocate memory\n", myname);
    quit(1);
  }
}

int machine_init (struct statics *statics)
{
  statics->procstate_names = procstatenames;
  statics->cpustate_names = cpustatenames;
  statics->memory_names = memorynames;
  statics->order_names = ordernames;
#ifdef COLOUR
  statics->colour_names = colournames;
#endif

  if (!(procdir = opendir(PROCFS))) {
    (void) fprintf(stderr, "Unable to open %s\n",PROCFS);
    return (-1);
  }
  
  if (chdir(PROCFS)) {
    (void) fprintf (stderr, "Unable to chdir to %s\n",PROCFS);
    return (-1);
  }
  
  return (0);
}

char *
format_header (register char *uname_field)
{
  register char *ptr;

  ptr = header + UNAME_START;
  while (*uname_field != '\0')
    *ptr++ = *uname_field++;

  return (header);
}

void
get_system_info (struct system_info *si)
{
  static kstat_ctl_t *kc = NULL;
  kstat_t *ks;
  kstat_named_t *kn;
  int i;
  static long cp_time[CPUSTATES];
  static long cp_old[CPUSTATES];
  static long cp_diff[CPUSTATES];
  cpu_stat_t cpu_stat;
  vminfo_t vminfo;
  int pagesize = sysconf(_SC_PAGESIZE);
  unsigned int time_delta;
  static vminfo_t vminfo_old;
  static hrtime_t snaptime_old = 0;
  
  if (!kc) {
    if (!(kc = kstat_open())) {;
      perror("kstat_open() ");
      quit(1);
    }
  }
  
  if (kstat_chain_update(kc) < 0) {
    perror("kstat_chain_update() ");
    quit(1);
  }

  si->last_pid = -1;   /* Ignore */
  
  ks = kstat_lookup(kc, "unix", 0, "system_misc");
  if (kstat_read(kc,ks,0) == -1) {
    perror("kstat_read() ");
    quit(1);
  }
  if (!(kn = kstat_data_lookup(ks, "avenrun_1min"))) {
    perror("kstat_data_lookup(avenrun_1min ");
    quit(1);
  }
  si->load_avg[0] = (double)kn->value.ui32 / FSCALE;
  if (!(kn = kstat_data_lookup(ks, "avenrun_5min"))) {
    perror("kstat_data_lookup(avenrun_5min) ");
    quit(1);
  }
  si->load_avg[1] = (double)kn->value.ui32 / FSCALE;
  if (!(kn = kstat_data_lookup(ks, "avenrun_15min"))) {
    perror("kstat_data_lookup(avenrun_15min) ");
    quit(1);
  }
  si->load_avg[2] = (double)kn->value.ui32 / FSCALE;

  /* On our first run through, use nproc to give a good
     starting value for the process array.
   */
  if (process_array_size == 0) {
    if (!(kn = kstat_data_lookup(ks, "nproc"))) {
      perror("kstat_data_lookup(nproc) ");
      quit(1);
    }
    realloc_arrays(kn->value.ui32);
  }

  /* Note: This seems to return the number of CPUs on
     the system, not the number of 'online' CPUS
  */
  if (!(kn = kstat_data_lookup(ks, "ncpus"))) {
    perror("kstat_data_lookup(ncpus) ");
    quit(1);
  }
  ncpus = kn->value.ui32;

  /* CPU/wait times */
  memset(cp_time,0,sizeof(cp_time));

#if 0    
  for (i = 0 ; i < ncpus ; i++) {
    ks = kstat_lookup(kc,"cpu_stat",i,NULL);
    if (!ks) {
      perror("kstat_lookup() ");
      quit(1);
    }
    if (kstat_read(kc,ks,&cpu_stat) < 0) {
      perror("kstat_read() ");
      quit(1);
    }
    cp_time[0] += (long) cpu_stat.cpu_sysinfo.cpu[0];       /* idle */
    cp_time[1] += (long) cpu_stat.cpu_sysinfo.cpu[1];       /* user */
    cp_time[2] += (long) cpu_stat.cpu_sysinfo.cpu[2];       /* kernel */
    cp_time[3] += (long) cpu_stat.cpu_sysinfo.wait[W_IO] +  /* io wait */
      (long) cpu_stat.cpu_sysinfo.wait[W_PIO];
    cp_time[4] += (long) cpu_stat.cpu_sysinfo.wait[W_SWAP]; /* swap wait */

  }
#else
  for (i = ncpus,ks = kc->kc_chain ; ks != NULL && i > 0 ; ks = ks->ks_next) {
    if (!strcmp(ks->ks_module,"cpu_stat")) {  
      i--;  
      if (kstat_read(kc,ks,&cpu_stat) < 0) {  
        perror("kstat_read() ");  
        quit(1);  
      }  
      cp_time[0] += (long) cpu_stat.cpu_sysinfo.cpu[0];       /* idle */  
      cp_time[1] += (long) cpu_stat.cpu_sysinfo.cpu[1];       /* user */  
      cp_time[2] += (long) cpu_stat.cpu_sysinfo.cpu[2];       /* kernel */  
      cp_time[3] += (long) cpu_stat.cpu_sysinfo.wait[W_IO] +  /* io wait */  
        (long) cpu_stat.cpu_sysinfo.wait[W_PIO];  
      cp_time[4] += (long) cpu_stat.cpu_sysinfo.wait[W_SWAP];
    } 
  } 
#endif
  (void) percentages(CPUSTATES,cpu_states,cp_time,cp_old,cp_diff);
  si->cpustates = cpu_states;

  /* Memory */
  memory_stats[0] = (sysconf(_SC_PHYS_PAGES) / 1024) * pagesize;

  if (!(ks = kstat_lookup(kc, "unix", 0, "vminfo"))) {
    perror("kstat_lookup() ");
    quit(1);
  }

  if (kstat_read(kc,ks,&vminfo) < 0) {
    perror("kstat_read() ");
    quit(1);
  }

  if (!snaptime_old) {
    vminfo_old.swap_resv = 0;
    vminfo_old.swap_avail = 0;
    vminfo_old.freemem = 0;
  }

  time_delta = ((ks->ks_snaptime-snaptime_old + 5e8) / 1e9);  
  if (time_delta) {
    if (vminfo.freemem > vminfo_old.freemem)
      memory_stats[1] = ((vminfo.freemem - vminfo_old.freemem) / time_delta)
	* (pagesize / 1024);
    if (vminfo.swap_resv > vminfo_old.swap_resv) 
      memory_stats[2] = ((vminfo.swap_resv - vminfo_old.swap_resv) / time_delta)
	* (pagesize / 1024);
    if (vminfo.swap_avail > vminfo_old.swap_avail)
      memory_stats[3] = ((vminfo.swap_avail - vminfo_old.swap_avail) / time_delta)
	* (pagesize / 1024);
  }
  snaptime_old = ks->ks_snaptime;
  vminfo_old.freemem = vminfo.freemem;
  vminfo_old.swap_resv = vminfo.swap_resv;
  vminfo_old.swap_avail = vminfo.swap_avail;

  si->memory = memory_stats;
}

old_proc_info_t *
get_old_proc (pid_t pid)
{
  old_proc_info_t *scan;

  if (!old_process_array) return NULL;
  
  for (scan = old_process_array ; scan->pr_valid != 0 ; scan++) {
    if (scan->pr_pid == pid)
      return scan;
  }
  return NULL;
}

caddr_t
get_process_info (struct system_info *si,struct process_select *sel,
		  int (*compare) ())
{
  int fd,proc_no,i;
  char filename_buf[32];
  psinfo_t psinfo;
  prusage_t psusage;
  struct dirent *direntp;
  proc_info_t *proc;
  old_proc_info_t *old_proc;
  static handle_t handle;
    
  memset(process_states,0,sizeof(process_states));
  si->p_total = 0;
  si->p_active = 0;
  
  for (proc_no = 0, rewinddir (procdir) ; (direntp = readdir (procdir)); ) {
    if (proc_no >= process_array_size)
      realloc_arrays(proc_no);
    proc = &process_array[proc_no];
    
    if (direntp->d_name[0] == '.')
      continue;

    snprintf(filename_buf,32,"%s/psinfo", direntp->d_name);
    if ((fd = open (filename_buf, O_RDONLY)) < 0)
      continue;

    if (read (fd, &psinfo, sizeof(psinfo_t)) != sizeof(psinfo_t)) {
      (void) close(fd);
      continue;
    }
    (void) close(fd);

    snprintf(filename_buf,32,"%s/usage", direntp->d_name);
    if ((fd = open (filename_buf, O_RDONLY)) < 0)
      continue;

    if (read (fd, &psusage, sizeof(prusage_t)) != sizeof(prusage_t)) {
      (void) close(fd);
      continue;
    }
    (void) close(fd);

    if (psinfo.pr_lwp.pr_state == 0)
      continue;

    si->p_total++;
    process_states[psinfo.pr_lwp.pr_state]++;
    
    /* Filter out the processes we want to see */
    /* idle */
    if (!sel->idle && psinfo.pr_pctcpu == 0 &&
	(psinfo.pr_lwp.pr_state != SRUN && psinfo.pr_lwp.pr_state != SONPROC))
      continue;
    /* system */
    if (!sel->system && (psinfo.pr_flag & SSYS))
      continue;
    /* uid */
    if (sel->uid != -1 && psinfo.pr_uid != sel->uid) 
      continue;
    /* command */
    if (sel->command && strcmp(psinfo.pr_fname,sel->command))
      continue;

    /* Fill our proc_info struct with data from this process */
    proc->pr_nlwp   = psinfo.pr_nlwp;
    proc->pr_pid    = psinfo.pr_pid;
    proc->pr_uid    = psinfo.pr_uid;
    proc->pr_size   = psinfo.pr_size;
    proc->pr_rssize = psinfo.pr_rssize;
    proc->pr_pctcpu = psinfo.pr_pctcpu;
    proc->pr_time   = psinfo.pr_time;
    strncpy(proc->pr_fname,psinfo.pr_fname,PRFNSZ);
    proc->pr_onpro  = psinfo.pr_lwp.pr_onpro;
    proc->pr_pri    = psinfo.pr_lwp.pr_pri;
    proc->pr_nice   = psinfo.pr_lwp.pr_nice;
    proc->pr_state  = psinfo.pr_lwp.pr_state;
    /* need to find the last sample to get the flts */
    if ((old_proc = get_old_proc(proc->pr_pid))) 
      proc->pr_majf = psusage.pr_majf - old_proc->pr_majf_abs;
    else
      proc->pr_majf = 0;
    proc->pr_majf_abs = psusage.pr_majf;

    si->p_active++;    
    /* Successfully added entry, so increment proc_no */
    proc_no++;
  }
  /* We now have an array of proc_no elements of 'active' processes
   * Scan through this and recreate the old_info array.
   */
  memset(old_process_array, 0, process_array_size * sizeof(old_proc_info_t));

  for (i = 0 ; i < proc_no ; i++) {
    old_process_array[i].pr_valid = 1;
    old_process_array[i].pr_pid = process_array[i].pr_pid;
    old_process_array[i].pr_majf_abs = process_array[i].pr_majf_abs;
  }
				      
  si->procstates = process_states;

  if (compare != NULL)
    qsort (process_array, proc_no, sizeof(proc_info_t), compare);

  handle.current_index = 0;
  
  return (caddr_t)&handle;
}

/* comparison routines for qsort */

/*
 * There are currently four possible comparison routines.  main selects
 * one of these by indexing in to the array proc_compares.
 *
 * Possible keys are defined as macros below.  Currently these keys are
 * defined:  percent cpu, cpu ticks, process state, resident set size,
 * total virtual memory usage.  The process states are ordered as follows
 * (from least to most important):  WAIT, zombie, sleep, stop, start, run.
 * The array declaration below maps a process state index into a number
 * that reflects this ordering.
 */

/* First, the possible comparison keys.  These are defined in such a way
   that they can be merely listed in the source code to define the actual
   desired ordering.
 */

#define ORDERKEY_PCTCPU  if ((result = p2->pr_pctcpu - p1->pr_pctcpu) == 0)

#define ORDERKEY_CPTICKS if ((result = p2->pr_time.tv_sec - p1->pr_time.tv_sec) == 0)
#define ORDERKEY_STATE   if ((result = (long) (sorted_state[p2->pr_state] - \
			       sorted_state[p1->pr_state])) == 0)
#define ORDERKEY_PRIO    if ((result = p1->pr_pri - p2->pr_pri) == 0)
#define ORDERKEY_RSSIZE  if ((result = p2->pr_rssize - p1->pr_rssize) == 0)
#define ORDERKEY_MEM     if ((result = (p2->pr_size - p1->pr_size)) == 0)
#define ORDERKEY_FLTS    if ((result = (p2->pr_majf - p1->pr_majf)) ==0 )
			     
/* Now the array that maps process state to a weight */

unsigned char sorted_state[] =
{
  0,				/* not used		*/
  3,				/* sleep		*/
  6,				/* run			*/
  2,				/* zombie		*/
  4,				/* stop			*/
  5,				/* start		*/
  7,				/* run on a processor   */
  1				/* being swapped (WAIT)	*/
};

/* compare_cpu - the comparison function for sorting by cpu percentage */
int
compare_cpu (proc_info_t *p1, proc_info_t *p2)
{
  register long result;
  
  ORDERKEY_PCTCPU
    ORDERKEY_CPTICKS
    ORDERKEY_STATE
    ORDERKEY_PRIO
    ORDERKEY_RSSIZE
    ORDERKEY_MEM
    ORDERKEY_FLTS
    ;

  return (result);
}

/* compare_size - the comparison function for sorting by total memory usage */
int
compare_size (proc_info_t *p1, proc_info_t *p2)
{
  register long result;
  
  ORDERKEY_MEM
    ORDERKEY_RSSIZE
    ORDERKEY_PCTCPU
    ORDERKEY_CPTICKS
    ORDERKEY_FLTS
    ORDERKEY_STATE
    ORDERKEY_PRIO
    ;
  
  return (result);
}

/* compare_res - the comparison function for sorting by resident set size */
int
compare_res (proc_info_t *p1, proc_info_t *p2)
{
  register long result;
  
  ORDERKEY_RSSIZE
    ORDERKEY_MEM
    ORDERKEY_PCTCPU
    ORDERKEY_CPTICKS
    ORDERKEY_FLTS
    ORDERKEY_STATE
    ORDERKEY_PRIO
    ;
  
  return (result);
}

/* compare_time - the comparison function for sorting by total cpu time */
int
compare_time (proc_info_t *p1, proc_info_t *p2)
{
  register long result;
  
  ORDERKEY_CPTICKS
    ORDERKEY_PCTCPU
    ORDERKEY_STATE
    ORDERKEY_PRIO
    ORDERKEY_MEM
    ORDERKEY_RSSIZE
    ORDERKEY_FLTS
    ;
  
  return (result);
}

/* compare_flts - the comparison function for sorting by major faults */
int
compare_flts (
	      proc_info_t *p1,
	      proc_info_t *p2)
{
  register long result;

  ORDERKEY_FLTS
  ORDERKEY_PCTCPU
  ORDERKEY_CPTICKS
  ORDERKEY_STATE
  ORDERKEY_PRIO
  ORDERKEY_RSSIZE
  ORDERKEY_MEM
    ;
  
  return (result);
}


char *
format_next_process (caddr_t ptr, char *(*get_userid) ())
{
  static char fmt[MAX_COLS];
  char state_buf[16];
#ifdef COLOUR
  char cpu_colour[32],flts_colour[32],size_colour[32],time_colour[32],
    res_colour[32];
#endif
  handle_t *handle = (handle_t *)ptr;
  proc_info_t *p = &process_array[handle->current_index++];
  double cpu_usage;

  if (p->pr_state == SONPROC) {
    if (p->pr_onpro < 100) 
      snprintf(state_buf, sizeof(state_buf), "cpu%02d", p->pr_onpro);
    else
      strncpy(state_buf, "cpuxx", sizeof(state_buf));
  } else
    strncpy(state_buf, state_abbrev[p->pr_state], sizeof(state_buf));

  cpu_usage = ((double)p->pr_pctcpu/0x8000)*100;
  
  snprintf(fmt,MAX_COLS,Proc_format,
	   p->pr_pid,
	   (*get_userid) (p->pr_uid),
	   p->pr_nlwp>999?999:p->pr_nlwp,
	   p->pr_pri>99?99:p->pr_pri,
	   p->pr_nice - NZERO,
#ifdef COLOUR
	   get_colour("size",p->pr_size/1024,size_colour,sizeof(size_colour)),
	   format_k(p->pr_size),
	   stop_colour(),
#else
	   format_k(p->pr_size),
#endif
#ifdef COLOUR
	   get_colour("res",p->pr_rssize/1024,res_colour,sizeof(res_colour)),
	   format_k(p->pr_rssize),
	   stop_colour(),
#else
	   format_k(p->pr_rssize),
#endif
	   state_buf,
#ifdef COLOUR
	   get_colour("time",p->pr_time.tv_sec,time_colour,
		      sizeof(time_colour)),
	   format_time(p->pr_time.tv_sec),
	   stop_colour(),
#else
	   format_time(p->pr_time.tv_sec),
#endif
#ifdef COLOUR
	   get_colour("flts",p->pr_majf,flts_colour,
		      sizeof(flts_colour)),
	   p->pr_majf,
	   stop_colour(),
#else
	   p->pr_majf,
#endif
#ifdef COLOUR
	   /* On solaris each thread can only get 100%/NCPUS usage, so
	      we up this to allow the TOPCOLOURS env var to give the
	      same results on single or multi processor machines */
	   get_colour("cpu",(int)cpu_usage*ncpus,cpu_colour,
		      sizeof(cpu_colour)),
	   cpu_usage,
	   stop_colour(),
#else
	   cpu_usage,
#endif
	   p->pr_fname);

#ifdef COLOUR_DEBUG
  if (strlen(fmt) > MAX_COLS-20) {
    printf("Borderline on MAX_COLS (%d)\n",strlen(fmt));
    quit(1);
  }
#endif
	   
  return (fmt);
}

/*
 * proc_owner(pid) - returns the uid that owns process "pid", or -1 if
 *              the process does not exist.
 *
 *  We don't store all processes in memory, so we open up the /proc
 *     info file.  Not as nice, but this should only be used rarely.
*/
uid_t
proc_owner (pid_t pid)
{
  char buf[30];
  int fd;
  psinfo_t psinfo;
  
  snprintf(buf, sizeof(buf),"%s/%d/psinfo", PROCFS, pid);
  if ((fd = open(buf, O_RDONLY)) < 0)
    return (-1);

  if (read(fd, &psinfo, sizeof(psinfo_t)) != sizeof(psinfo_t)) {
    (void) close(fd);
    return (-1);
  }
  (void) close(fd);
  return (psinfo.pr_uid);
}
